/*

  pic18.c
        Load PIC18F series parts using the PIC-EL hardware.
        dvconfig.txt has PIC device information
                in theory, add a device to this file and it works

        Written in Mix Power C
        (c) 2004  Ron Carr

*/


#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <malloc.h>

/* defines */

/* delays, anything under 1us becomes 1us, others 1.5 min value */
/* longest delay we can do with this setup is 32ms */
#define P2 1
#define P3 1
#define P4 1
#define P5 1
#define P6 1
#define P9 1500
#define P10 8
#define P11 8000
#define P12 3
#define P13 1
#define P14 1
#define P15 3

/* rs232 */
/* break in lcr, rts dtr in mcr, cts in msr */
#define LCR 3
#define MCR 4
#define MSR 6
#define RTS 2
#define DTR 1
#define CTS 0x10
#define BREAK 0x40

/* global vars */

FILE *fp;
unsigned int dc = 128;   /* delay counter 1us */

struct device{
 char name[10];
 int  id2;
 int  id1;
 int  codesize;
 int  eesize;
 int  configsize;
 int  bcount;       /* block count */
}device[10];

struct block{
 char name[12];
 int  start;      /* start address */
 int  end;
 int  prog;       /* need to program */
 int  erase_cmd;  /* command to erase this block */
 }block[5];

int devcount;

int configmasks[20];  /* all seem to be 14, load active masks only */

struct device *pic;   /* active device */
char hexfile[40];     /* 8.3 filename, room for some path */
int combase;
int rtsi, ctsi, dtri, tdi;  /* signal invert bits */

int configok;   /* verifys as is before programming */
int eepromok;
int progcount;  /* bytes written */
int errors;
int mindelay= 1;  /* read from config file */

char far * hexd;  /* hex file data */
char far * dutd;  /* data read from part */

/* all 18f devices in datasheet have 256 byte eeprom */
char eehexd[256];
char eedutd[256];
char eetouch[256];   /* only change ee locations specified in hex file */

/* config data, configsize bytes used */
char cfhexd[20];
char cfdutd[20];

int main( int argc, char*argv[] ){
char ans[256];

   printf("\n *** pic18.exe   (c)2004  Ron Carr *** \n\n");

   cleanup();  /* init hardware to off */

   /* get name of hex file to load */
   if( argc > 1 ) strncpy( hexfile,argv[1],35 );
   else{
      printf("Enter the name of the hex file ");
      gets(hexfile);
      printf("\n\n");
      }

   if ( strchr(hexfile,'?') > 0 ){
      printf("Usage\npic18 HexFileName\n\n");
      exit(1);
      }

   if( strlen(hexfile) == 0 ) ;  /* ok, erase only */
   else if( strstr(hexfile,".hex") == 0 ) strcat(hexfile,".hex");

 
   calc_delay(); /* before getting value for mindelay from dconfig */

   if (read_device_file()){
      printf("Missing config file dconfig.txt");
      exit(1);
      }
   dump_dev_table();    /* useful to see what went wrong when device not recognized */

   /* start accessing hardware */
   dev_reset(1);
   pic= read_config();

   /* try 2nd reset if failed */
   if( pic == 0 ){
      printf("\n");
      dev_reset(0);
      pic= read_config();
      }

   if( pic == 0 ){
      printf("Device not recognized");
      cleanup();
      exit(1);
      }

   read_masks();
   make_blocks();

   allo_buffers();

   if( load_hex() == 0 ){   /* offer to bulk erase */
      printf("Do you want to bulk erase this device ");
      gets( ans );
      if( strcmpi(ans,"yes") == 0 ){
         printf("bulk erasing ..." );
         erase(0x80);
         verify(1);
         }
      }
   else{      /* normal programming */
      verify(0);
      erase_if_needed();
      prog_code();
      prog_eeprom();
      prog_config();
      verify(1);
      if( errors == 0 ) printf("Passed\n");
      else printf("Failed\n");
      }

   /* all done */
   cleanup();
   farfree( hexd );
   farfree( dutd );

}

/* pick up two characters in the hex file and sum up checksum */
char atox( char **a, char* chk){
char work[5];
char val;
   
    strcpy(work,"0x");
    work[2]= **a;  (*a)++;   /* also increment the pointer */
    work[3]= **a;  (*a)++;
    work[4]= 0;
    /* printf("%s ",work); */
    sscanf(work,"%i",&val);
    *chk+= val;
    return val;
}

void allo_buffers(){

int i;

   hexd= farmalloc( (unsigned long)pic->codesize );
   dutd= farmalloc( (unsigned long)pic->codesize );
   if( hexd == 0 || dutd == 0 ) {
      printf("Out of memory in far heap\n");
      exit(1);
      }

   /* initialize to 0xff, erased state */
   farmemset( hexd,0xff,(size_t)pic->codesize );
   farmemset( dutd,0xff,(size_t)pic->codesize );

   /* init ee and config buffers too */
   memset(eehexd,0xff,sizeof(eehexd));
   memset(eedutd,0xff,sizeof(eedutd));
   memset(eetouch,0x00,sizeof(eetouch));

   for( i= 0; i < pic->configsize; ++i ){
      cfhexd[i]= cfdutd[i]= configmasks[i];
      }

}


/* read the hex file and store in data storage previous allocated */
/* set flags to program data in hex file */
int load_hex(){
char bcount;
unsigned int adr;
char type;
char data;
char check;
char line[256];
char *p;
char byte;
unsigned int extended;

   fp= fopen(hexfile,"r");
   if( fp == 0 ){
      printf("No hex data to load !\n");
      return 0;
      }
   while( p=fgets(line,254,fp) ){
      /* header stuff */
      if( *p != ':' ) continue;
      p++;  /* skip : */
      check= 0;
      bcount= atox(&p,&check);
      byte = atox(&p,&check); 
      adr= byte << 8;
      byte = atox(&p,&check);
      adr|= byte;
      type= atox(&p,&check);
      switch( type ){
         case 1:   /* end of file */
         byte = atox(&p,&check);
         fclose(fp);
         if( check != 0 ){
            printf("Checksum error in hex file\n");
            return 0;
            }
         return 1;
         break;

         case 4:   /* extended address */
         if( bcount != 2 ){
            printf("Unexpected size of extended address in hex file\n");
            fclose(fp);
            return 0;
            }
         byte = atox(&p,&check);
         extended= byte << 8;
         byte = atox(&p,&check);
         extended|= byte;
         byte = atox(&p,&check);
         if( check != 0 ){
            printf("Checksum error in hex file\n");
            fclose(fp);
            return 0;
            }
         break;

         case 0:   /* regular data */
         while( bcount--){
            byte = atox(&p,&check);

            switch(extended){
            case 0:  /* code */
            if( adr < pic->codesize ) *(hexd + adr) = byte;
            bufprog(adr);
            ++adr;
            break;

            case 0x0020:   /* ignoreing user id bytes */
            break;

            case 0x0030:   /* config */
            if( adr < pic->configsize ) cfhexd[adr]= byte & configmasks[adr];
            ++adr;
            break;

            case 0x00f0:  /* eeprom */
            if( adr < 256 ) {
               eetouch[adr] = 1;
               eehexd[adr++]= byte;
               }

            break;
            }/* end switch */

            } /* end while bcount */
         byte= atox(&p,&check);
         if( check != 0 ){
            printf("Checksum error in hex file\n");
            fclose(fp);
            return 0;
            }         
         break;  /* end case 0 */
         } /* end switch */
      }/* end while */
   fclose(fp);
   printf("No end of record in hex file\n");
   return 0;
}

/* flag a block for programming */
/* would be nice if this was outside the innermost loop */
/* marking blocks for programming allows preserving contents
   of blocks not mentioned in the hex file */
void bufprog( int adr ){
int i;

    for( i= 0; i <= pic->bcount; ++i ){
       if( block[i].start <= adr && block[i].end >= adr ){
          block[i].prog= 1;
          break;
          }
       }
}


void verify( int report ){

int i;
int val;
int b;

   /* read each section and compare to hex file data */
   /* report errors if report */
   /* set/clear block[].prog if no report */

   if( report ) printf("Verify...\n");

   /* eeprom */
   eepromok = 1;
   for( i= 0; i < 256; ++i ){
      cmd(0,0x9ea6);
      cmd(0,0x9ca6);
      cmd(0,(0x0e << 8 ) | i );
      cmd(0,0x6ea9);
      cmd(0,0x80a6);
      cmd(0,0x50a8);
      cmd(0,0x6ef5);
      val= readee();

      eedutd[i]= (char)val;
      if( eetouch[i]  && val != eehexd[i] ){
         if ( report ){
            printf("eeprom verify failed at %d\n",i);
            printf("Expected %x, Found %x\n",eehexd[i],val);
            ++errors;
            break;
            }
         else eepromok= 0;
         }
      }/* end eeprom */

   /* config */
   configok= 1;
 /*  cmd(0,0x8ea6);
   cmd(0,0x8ca6);
 */
   cmd(0,0x0e30);
   cmd(0,0x6ef8);
   cmd(0,0x0e00);
   cmd(0,0x6ef7);
   cmd(0,0x0e00);
   cmd(0,0x6ef6);
   for( i= 0; i < pic->configsize; ++i ){
      val= readpic();
      /* printf("%2x ",val); */
      cfdutd[i]= (char)(val & configmasks[i]);
      if( cfdutd[i] != cfhexd[i] ){
         if( report ){
            printf("config byte verify failed at %d\n",i);
            printf("Expected %x, Found %x\n",cfhexd[i],cfdutd[i]);
            ++errors;
            break; 
            }
         else configok= 0;
         }
      }   /* end config check */

   /* read code by block */
   /* clear prog if match */
   for( b= 0; b < 5; ++b ){

      if( block[b].prog ){

         block[b].prog= 0;   /* assume ok */
   /*
         cmd(0,0x8ea6);
         cmd(0,0x9ca6);
   */
         cmd(0,0x0e00);
         cmd(0,0x6ef8);
         cmd(0,( 0x0e << 8 ) | ( block[b].start >> 8 ) & 0xff );
         cmd(0,0x6ef7);
         cmd(0,( 0x0e << 8 ) | ( block[b].start & 0xff ) );
         cmd(0,0x6ef6);
         
         for( i= block[b].start; i < block[b].end; ++i ){
            val= readpic() & 0xff;
            /* printf("%x ",val); */
            *(dutd + i )= (char)val;
            if( *(hexd + i ) != val ){
               if( report ){
                  printf("Code verify failed at %d\n", block[b].start + i);
                  printf("Expected %x, Found %x\n",*(hexd + i),val);
                  ++errors;
                  break; 
                  }
               else block[b].prog= 1;  /* turn back on */
               }
            }
         } /* end block */
      } /* end code verify */

}

int readee(){
int val;
int i;
int tmp;

   CLK(1,0);  
   CLK(0,0);
   CLK(1,1);  
   CLK(0,1);
   CLK(1,0);  
   CLK(0,0);
   CLK(1,0);
   CLK(0,0);
   delay(P5);

   for(i = 0; i < 8; ++i ){
      CLK(1,0);
      CLK(0,0);
      }
   delay(P6);

   /* need sent data high due to wired or */
   CLK(0,1);
   /* data now avail, read 8 bits */
   val= 0;
   for(i= 1; i < 256; i <<= 1 ){
      /* data sheet shows data changing with high clock */
      CLK(1,1);        /* pause(); */
      CLK(0,1);
      tmp= inportb(combase+MSR) & CTS;
      if( ctsi ) tmp ^= CTS;   /* inverted data */
      val= val | ( ( tmp )  ? i : 0 );
      }
   delay(P5);
   return val;

}



void erase( int erase_command ){

   if( erase_command == 0 ) return;

   cmd(0,0x0e3c);
   cmd(0,0x6ef8);
   cmd(0,0x0e00);
   cmd(0,0x6ef7);
   cmd(0,0x0e04);
   cmd(0,0x6ef6);
   cmd(0xc,erase_command);
   cmd(0,0);
   cmd11();
  

}

void erase_if_needed(){

int i,k;

   for( i= 0; i < 5; ++i ){
      if( block[i].prog == 0 ) continue;
      printf("Erasing %s \n",block[i].name);
      erase(block[i].erase_cmd);
      /* clear dut data */
      farmemset( dutd + block[i].start,0xff,
           (size_t)(block[i].end - block[i].start + 1));
      }

}


void prog_code(){

int i;
int adr;

   for( i= 0; i < 5; ++i ){
      if( block[i].prog ){
         cmd(0,0x8ea6);     /* allow access to code mem */
         cmd(0,0x9ca6);
         progcount= 0;
         printf("Programming %s...",block[i].name);
         for( adr= block[i].start; adr < block[i].end; adr+= 8 )
               prog_c8( adr );
         printf(" %d locations\n",progcount);
         }
      }

}


/* program 8 locations at adr */
void prog_c8( int adr ){

char rdata[8];
char wdata[8];
int i;
int match;

    /* get data local for easy access */
   for(i= 0; i < 8; ++i ){
      rdata[i] = *( dutd + adr + i );
      wdata[i] = *( hexd + adr + i );
      }

    /* see if buffers match, ie don't need to program */

   match= 1;
   for( i = 0; i < 8; ++i ){
      if( rdata[i] != wdata[i] ) match= 0;
      }
   if( match == 1 ) return;

    /* program 8 locations */

   cmd(0,0x0e00);   /* may want long adr at some point */
   cmd(0,0x6ef8);
   cmd(0,( 0x0e << 8 ) | (( adr >> 8 ) & 0xff) );
   cmd(0,0x6ef7);
   cmd(0,(0x0e << 8 ) | (adr & 0xff ) );
   cmd(0,0x6ef6);
   cmd(0xd,(wdata[1] << 8) | wdata[0]);
   cmd(0xd,(wdata[3] << 8) | wdata[2]);
   cmd(0xd,(wdata[5] << 8) | wdata[4]);
   cmd(0xf,(wdata[7] << 8) | wdata[6]);
   cmd9();

   progcount+= 8;

}


void prog_eeprom(){
int i;

   if( eepromok ){
      printf("EEprom verified as OK\n");
      return;
      }
   printf("Programming EEPROM...");
   progcount= 0;
   for( i= 0; i < 256; ++i ){
      if( eehexd[i] == eedutd[i] ) continue;
      if( eetouch[i] == 0 )continue;    /* only change data in hex */
      cmd(0,0x9ea6);
      cmd(0,0x9ca6);
      cmd(0,( 0x0e << 8 ) | i );
      cmd(0,0x6ea9);
      cmd(0,( 0x0e << 8 ) | eehexd[i] );
      cmd(0,0x6ea8);
      cmd(0,0x84a6);
      cmd(0,0x0e55);
      cmd(0,0x6ea7);
      cmd(0,0x0eaa);
      cmd(0,0x6ea7);
      cmd(0,0x82a6);
      cmd(0,0);
      cmd11();
      cmd(0,0x94a6);
      progcount+= 1;
      }
   printf("%d locations\n",progcount);
}

void prog_config(){
int i;
char lsb,msb;

   if( configok ){
      printf("Config bits verified as OK\n");
      return;
      }
   progcount= 0;
   printf("Programming Config...");
   for(i= 0; i < pic->configsize; i+= 2){
      if( cfhexd[i] == cfdutd[i] && cfhexd[i+1] == cfdutd[i+1] ) continue;
      lsb= cfhexd[i];  msb= cfhexd[i+1];
      cmd(0,0x8ea6);
      cmd(0,0x8ca6);
      cmd(0,0xef00);
      cmd(0,0xf800);
      cmd(0,0x0e30);
      cmd(0,0x6ef8);
      cmd(0,0x0e00);
      cmd(0,0x6ef7);
      cmd(0,( 0x0e << 8 ) | i);   /* address */
      cmd(0,0x6ef6);
      cmd(0xf,(msb << 8) | lsb );  /* found msb,lsb swapped from documentation */
      cmd9();
      cmd(0,0x2af6);
      cmd(0xf,(msb << 8) | lsb );
      cmd9();
      progcount+= 1;
      }
   printf("%d locations\n",progcount);
}

cmd9(){
int i;

      CLK(1,0);
      CLK(0,0);
      CLK(1,0);
      CLK(0,0);
      CLK(1,0);
      CLK(0,0);
      CLK(1,0);
      delay(P9);
      CLK(0,0);
      delay(P10);

   for( i= 0; i < 16; ++i ){
      CLK(1,0 );
      CLK(0,0 );
      }

}

cmd11(){
int i;

      CLK(1,0);
      CLK(0,0);
      CLK(1,0);
      CLK(0,0);
      CLK(1,0);
      CLK(0,0);
      CLK(1,0);
      CLK(0,0);
      delay(P11);
      delay(P10);

   for( i= 0; i < 16; ++i ){
      CLK(1,0 );
      CLK(0,0 );
      }

}
/* set up the block structure */
void make_blocks(){

int i;
int sob;

   /* we always have a boot block, but other blocks can be diffent sizes */
   sob= pic->codesize / pic->bcount;

   strcpy(block[0].name,"Boot Block");
   block[0].start= 0;
   block[0].end  = 0x1FF;
   block[0].erase_cmd = 0x83;

   for( i= 1; i < 5; ++i ){
      sprintf(block[i].name,"Block %d",i - 1);
      block[i].start= (i == 1 )? 0x200 : (i - 1) * sob ;
      block[i].end = i * sob - 1;
      block[i].erase_cmd = 0x87 + i;

      if( i * sob  >= pic->codesize ) break;
      }

}

/* reset some hardware */
void cleanup(){

   CLK(0,0);
   vpgm(0);

}



/*  read the config file again and get masks for the active device */
/*  store in configmasks[] */
void read_masks(){

char line[256];
char junk[256];
int i;

   if( pic->configsize > 20 ){
      printf("Error - config table too large for program\n");
      exit(1);
      }
   fp= fopen("dconfig.txt","r");

   while( fgets(line,254,fp)){
      if( line[0] == ';' ) continue;
      if( stristr(line,pic->name) != 0 ){
         for( i= 0; i < pic->configsize; ++i ){
            fgets(line,254,fp);
            sscanf(line,"%i",&configmasks[i]);
            }
         break;
         }
      }
  fclose(fp);

/*
  for( i = 0; i < pic->configsize; ++i ){
     printf("%2x\n",configmasks[i]);
     }
*/

}



struct device * read_config(){
/* reads config bits and finds device in table */
/* returns pointer to struct in use */
int i;
int id1,id2;

   /* set up to read the config */
   cmd(0,0x0e3f);
   cmd(0,0x6ef8);
   cmd(0,0x0eff);
   cmd(0,0x6ef7);
   cmd(0,0x0efe);
   cmd(0,0x6ef6);

   id1 = readpic();
   id2 = readpic();

   /* id1 is masked to used bits */
   id1&= 0xe0;

   printf("ID2:%2x ID1:%2x  ",id2,id1);  /* reverse as in datasheet */

   /* lookup devices in table and find a match */
   for( i= 0; i < devcount; ++i ){
      if( device[i].id1 == id1 && device[i].id2 == id2 ){
         printf("  Found %s\n",device[i].name);
         return &device[i];
         }
      }

   /* bad return */
   return (void *) 0;
}

int readpic(){  /* issue read table latch command */
int val;
int i;
int tmp;

   CLK(1,1);  /* pause(); */
   CLK(0,1);
   CLK(1,0);  /* pause(); */
   CLK(0,0);
   CLK(1,0);   /* pause(); */
   CLK(0,0);
   CLK(1,1);   /* pause(); */
   CLK(0,1);
   delay(P5);

   for(i = 0; i < 8; ++i ){
      CLK(1,0);
      CLK(0,0);
      }
   delay(P6);

   /* need sent data high due to wired or */
   CLK(0,1);
   /* data now avail, read 8 bits */
   val= 0;
   for(i= 1; i < 256; i <<= 1 ){
      /* data sheet shows data changing with high clock */
      CLK(1,1);        /* pause(); */
      CLK(0,1);
      tmp= inportb(combase+MSR) & CTS;
      if( ctsi ) tmp ^= CTS;   /* inverted data */
      val= val | ( ( tmp )  ? i : 0 );
      }
   delay(P5);
   return val;

}

void pause(){  /* debug aid */
char junk[30];
   
   printf("Hit return");   gets(junk);
}

dump_dev_table(){  /* debug aid */
int i;

  printf("Combase address %x\n",combase);
  for( i= 0; i < devcount; ++i ){
     pic= &device[i];
     printf("%s %x %x %x %x %d %d\n",
        pic->name,
        pic->id2,
        pic->id1,
        pic->codesize,
        pic->eesize,
        pic->configsize,
        pic->bcount );
     }
  printf("\n\n");

}

int read_device_file(){  /* read dconfig.txt */
char line[256];
char junk[256];

   if( (fp= fopen("dconfig.txt","r")) == 0) return 1;
   while( fgets(line,254,fp)){
      /* looking for a device line */
      if( line[0] == ';' ) continue;
      if( line[0] == '1'  && line[1] == '8' && toupper(line[2]) == 'F' ){
         loaddev(line);
         }
      if( stristr(line,"combase") != 0){
         sscanf(line,"%s %i %d %d %d %d"
               ,junk,&combase,&rtsi,&ctsi,&dtri,&tdi);
         }
      if( stristr(line,"mindelay") != 0){
         sscanf(line,"%s %d",junk,&mindelay);
         }

      }

   fclose(fp);   /* get config mask values later when we know what device */
   return 0;

}

void loaddev(char *p){


   if( devcount > 9 ) {
      printf("Too many devices defined in dconfig file\n");
      return;
      }

   /* i hate scanf but here goes */
   sscanf(p,"%s %i %i %i %i %d %d", &device[devcount].name,
          &device[devcount].id2,
          &device[devcount].id1,
          &device[devcount].codesize,
          &device[devcount].eesize,
          &device[devcount].configsize,
          &device[devcount].bcount );

   devcount+= 1;

}


void calc_delay(){  /* determine dc value for this machine for 1us */

/* binary converg to correct value */
int dir = 1;
int val;
clock_t tick;
int diff;
int outcount= 0;

   printf("Calculating delay value...");
   val= dc/2;
   while( val ){
      tick= clock();
      delay( 20000);  /* delay 10 ticks, 100ms */
      delay( 20000);
      delay( 20000);
      delay( 20000);
      delay( 20000);
      diff= (int)(clock() - tick);

      if( diff < 10 ){   /* too fast */
         if( dir == -1 ) {
            dir= 1;  val=  val/2;
            }
         dc= dc + val * dir; 
         }

      if( diff > 10 ){   /* too slow */
         if( dir == 1 ){
            dir= -1;  val=  val/2;
            }
         dc= dc + val * dir; 
         }

      if( dc < 0 ) dc= 0;
   /*   printf("%d,%d ",val*dir,dc); */
      if( val == 0 && ++outcount < 3 ) val= 1;
      }  /* end while */

      printf("%d\n\n",dc);
      /* add some slop for slow machine */
      dc+= 1;

}

void delay( count ){   /* delay one us per count */

int dc1;

   /* use mindelay parameter */
   if (count < mindelay ) count= mindelay;

   while ( count--){
      dc1= dc;
      while( dc1-- );
      }

}

void CLK( int hilo, int dat ){
/* clocks data to pic, delay */

char val;

   /* invert data and clock if needed */
   val= (dat ^ dtri)? DTR : 0;
   val= val | ((hilo ^ rtsi)? RTS : 0);
   outportb(combase+MCR,val);
   delay(P2);

}

void cmd( int command, unsigned int data ){
/* complete data/clock of one command */
int i;

   /* send command lsb - msb */
   for( i= 0; i < 4; ++i ){
      
      CLK(1, command & 1);
      CLK(0, command & 1);
      command >>= 1;
      }

   /* send 16 bit data */
   delay(P5);
   for( i= 0; i < 16; ++i ){
      CLK(1,data & 1 );
      CLK(0,data & 1 );
      data >>= 1;
      }
   delay(P5);
}

void dev_reset(int pause){  /* reset pic */

char junk[256];


   /* set clock and data low */
   CLK(0,0);
   vpgm(0);
   delay(P13);  

   if( pause ){
      printf("Enable VPP switch and power to device, hit return to cont...");
      gets(junk);
      }

   vpgm(1);  /* turn on 12 volts */

   delay(P12);


}

/* control 12v supply */
void vpgm( int val ){

char td;

   /* invert if required, use 8 bit char ( const 3 ) */
   td= ( val ^ tdi ) ? BREAK + 3 : 3;

   outportb(combase+LCR,td);

}

